// ignore_for_file: public_member_api_docs, non_constant_identifier_names, omit_local_variable_types, use_setters_to_change_properties, lines_longer_than_80_chars, avoid_positional_boolean_parameters, avoid_multiple_declarations_per_line, depend_on_referenced_packages

import 'dart:math';

import 'package:collection/collection.dart';

class AdaptiveHR {
  AdaptiveHR(int sps) {
    initHrVariable();
    movingFiltered.clear();
    lpFiltered.clear();

    if (sps == 0) {
      // 250sps
      LP_WINDOW = 37;
      fs = 250;
    } else if (sps == 1) {
      // 500sps
      LP_WINDOW = 74;
      fs = 500;
    }
    MAX_LIST_LEN = (fs * 2).toInt();
    SEC4 = (fs * 4).toInt();

    detectR = DetectPeaks(sps);
  }

  int rInd = 0;
  int r_detected = 0;
  int r_prev = 0;
  int hrB = 0;
  static double fs = 250;
  static int LP_WINDOW = 37;
  static int MAX_LIST_LEN = 500;
  static int SEC4 = 1000;
  final List<double> hrList = [];
  int notDetectedCnt = 0;
  int notDetectedCnt2 = 0;
  bool searchBool = false;

  double delta = 0;
  ExcludePeaks removeR = ExcludePeaks();
  DetectPeaks detectR = DetectPeaks(0);

  final List<double> dataList = [];
  final List<double> movingFiltered = [];
  final List<double> lpFiltered = [];

  void initHrVariable() {
    notDetectedCnt = 0;
  }

  double calculateMedian(List<double> data) {
    double medVal = 0;

    final copyData = List<double>.from(data);
    if (copyData.length.isOdd) {
      medVal = copyData[(copyData.length - 1) ~/ 2];
    } else {
      medVal = (copyData[copyData.length ~/ 2 - 1] + copyData[copyData.length ~/ 2]) / 2;
    }
    return medVal;
  }

  int calculateHR(
    double input,
    bool targetLeadState,
  ) {
    // target_leadState : false --> detached
    var hr_result = -1; // -1 : peak is not detected
    double output1 = 0, output2 = 0;

    if (targetLeadState) {
      dataList.add(input);
      if (dataList.length > LP_WINDOW) {
        dataList.removeAt(0);
      }

      output1 = removeR.movingAvFilt(input);
      movingFiltered.add(output1);
      output2 = removeR.lowpassFilt(output1);
      lpFiltered.add(output2);

      if (lpFiltered.length > MAX_LIST_LEN) {
        rInd = detectR.findPeaks(lpFiltered[lpFiltered.length - 1]);

        if (rInd > 0) {
          searchBool = true;
        }

        if (searchBool) {
          searchBool = false;
          notDetectedCnt = 0;

          rInd = detectR.adjustPeaks(dataList, rInd);

          //calculate HR
          r_detected = rInd;
          if (r_prev > 0) {
            hrB = 60.0 * fs ~/ (r_detected - r_prev);
            // hr_result = hrB;
          }

          r_prev = r_detected;

          hrList.add(hrB.toDouble());

          hr_result = calculateMedian(hrList).toInt();

          if (hrList.length >= 5) {
            hrList.removeAt(0);
          }
        } else {
          notDetectedCnt++;
          if (notDetectedCnt > SEC4) {
            // 1000sample = 4sec
            notDetectedCnt = 0;
            hr_result = 0;
          }
        }
      } else if (lpFiltered.length == MAX_LIST_LEN) {
        delta = detectR.getMedian(lpFiltered.sublist(0, MAX_LIST_LEN)) + detectR.meanAbsoluteDeviation(lpFiltered.sublist(0, MAX_LIST_LEN));
        detectR.inputDelta(delta);
      }
    } else {
      // lead is not attached
      notDetectedCnt2++;
      if (notDetectedCnt2 > fs) {
        // 1 sec
        // 1sec
        hr_result = 0;
        notDetectedCnt2 = 0;
      } else {
        hr_result = -1;
      }
    }

    return hr_result;
  }
}

class ExcludePeaks {
  final List<double> movingBuffer = [];
  final List<double> lpBuffer = [];
  final int movingWindow = 7;

  double thres1 = 0;
  double thres2 = 0;

  double sqData = 0;
  double filteredData = 0;
  int cnt = 0;
  int ind = 0;
  int fs = AdaptiveHR.fs.toInt();
  int qrsWindow = AdaptiveHR.fs ~/ 50;
  int width = 0;

  bool removeBool = false;
  bool qrsBool = false;

  double movingAvFilt(double sample) {
    filteredData = 0;
    movingBuffer.add(sample);
    if (movingBuffer.length == movingWindow) {
      filteredData = movingBuffer[(movingWindow / 2).floor()] - 1 / movingWindow * movingBuffer.reduce((a, b) => a + b);
      movingBuffer.removeAt(0);
    }
    return filteredData;
  }

  double lowpassFilt(double sample) {
    filteredData = 0;
    sqData = pow(sample, 2).toDouble();
    lpBuffer.add(sqData);
    if (lpBuffer.length == AdaptiveHR.LP_WINDOW) {
      filteredData = lpBuffer.reduce((a, b) => a + b);
      lpBuffer.removeAt(0);
    }
    return filteredData;
  }

  double std(List<double> data) {
    double stdSum = 0;
    final double dataAv = data.average;
    final int length = data.length;
    for (final double sample in data) {
      stdSum += pow(sample - dataAv, 2);
    }
    final double std = sqrt(stdSum / length);
    return std;
  }

  List<double> getDifference(List<double> data, List<double> fData) {
    final diffData = <double>[];
    for (var i = 0; i < data.length; i++) {
      diffData.add(data[i] - fData[i]);
    }
    final diffDataAverage = diffData.average;
    final stdDiffData = std(diffData);
    thres1 = diffDataAverage + 3 * stdDiffData;
    thres2 = diffDataAverage - 3 * stdDiffData;
    return [thres1, thres2];
  }

  List<double> calculateThreshold(
    List<double> diffData,
    int startInd,
    int endInd,
  ) {
    final diffRange = diffData.sublist(startInd, endInd);

    final diffRangeAverage = diffRange.average;
    final stdDiffData = std(diffData);
    thres1 = diffRangeAverage + stdDiffData;
    thres2 = diffRangeAverage - stdDiffData;
    return [thres1, thres2];
  }

  bool removeResidualQrs(double diffEcg, double thres1, double thres2) {
    if (diffEcg > thres1 || diffEcg < thres2) {
      removeBool = true;
    } else {
      removeBool = false;
    }

    return removeBool;
  }

  int checkQrs(double diffEcg, int index, double thres1, double thres2) {
    width = 0;
    if (diffEcg > thres1 || diffEcg < thres2) {
      if (ind > 0) {
        if (ind - index < qrsWindow) {
          cnt += ind - index;
          ind = index;

          if (cnt > 0.04 * fs) {
            if (qrsBool == true) {
              width = ind - index;
            } else {
              width = cnt;
              qrsBool = true;
            }
          }
        } else {
          cnt = 0;
          ind = 0;
          qrsBool = false;
        }
      }
      ind = index;
    }
    return width;
  }
}

class DetectPeaks {
  DetectPeaks(int sps) {
    if (sps == 0) {
      MAXTAB_MIN = 50;
      MAXTAB_MAX = 90;
      MAXTAB_BUFF = 25;
      MAXTAB_CHECK_LEN = 24;
      MAX_BUFF_LEN = 500;
    } else if (sps == 1) {
      MAXTAB_MIN = 100;
      MAXTAB_MAX = 180;
      MAXTAB_BUFF = 50;
      MAXTAB_CHECK_LEN = 48;
      MAX_BUFF_LEN = 1000;
    }
  }

  List<List<double>> maxtab = [];

  //List mintab = [];
  //List tab = [];

  double mn = double.infinity;
  double mx = -double.infinity;

  bool lookForMax = true;
  bool firstTry = true;
  int mxPos = 0;
  int mnPos = 0;
  double delta = 0;
  int x = -1;
  int cnt = -1;
  int ignore = 0;

  List<double> buffer = [];

  int newInd = 0;
  int ind = 0;
  double maxVal = 0;
  int qCand = 0;
  int sCand = 0;

  double midVal = 0;

  double slope1 = 0;
  double slope2 = 0;

  int MAXTAB_MIN = 50;
  int MAXTAB_MAX = 90;
  int MAXTAB_BUFF = 25;
  int MAXTAB_CHECK_LEN = 24;

  int MAX_BUFF_LEN = 500;

  void inputDelta(double deltaVal) {
    delta = deltaVal;
  }

  double meanAbsoluteDeviation(List<double> data) {
    final absData = <double>[];
    double element = 0;
    final double dataAv = data.average;

    for (final sample in data) {
      element = sample - dataAv;
      absData.add(element.abs());
    }
    return absData.average;
  }

  double getMedian(List<double> data) {
    double medVal = 0;

    data.sort();
    if (data.length.isOdd) {
      medVal = data[(data.length - 1) ~/ 2];
    } else {
      medVal = (data[data.length ~/ 2 - 1] + data[data.length ~/ 2]) / 2;
    }
    return medVal;
  }

  double calculateSlope(List<double> data) {
    final diffData = <double>[];
    for (var i = 0; i < data.length - 1; i++) {
      diffData.add(data[i + 1] - data[i]);
    }
    return diffData.average;
  }

  int findPeaks(double signalNew) {
    x++;
    cnt++;

    buffer.add(signalNew);

    if (signalNew > mx) {
      mx = signalNew;
      mxPos = x;
    }

    if (signalNew < mn) {
      mn = signalNew;
      mnPos = x;
    }

    if (cnt > MAX_BUFF_LEN) {
      cnt = 0;
      delta = getMedian(buffer.sublist(x - MAX_BUFF_LEN, x + 1)) + meanAbsoluteDeviation(buffer.sublist(x - MAX_BUFF_LEN, x + 1));
      firstTry = true;
      //added
      mxPos = 0;
      mnPos = 0;
      mn = double.infinity;
      mx = -double.infinity;
    }

    if (firstTry) {
      if (signalNew < mx - delta) {
        maxtab.add([mxPos.toDouble(), mx]);
        mn = signalNew;
        mnPos = x;
        lookForMax = false;
        firstTry = false;
        cnt = 0;
        return 0; //mxPos;
      }

      if (signalNew > mn + delta) {
        mx = signalNew;
        mxPos = x;
        lookForMax = true;
        firstTry = false;
        cnt = 0;
      }

      return 0;
    } else {
      if (lookForMax) {
        if (signalNew < mx - delta) {
          if (maxtab.isNotEmpty) {
            if (mxPos - maxtab.last[0] <= MAXTAB_MIN) {
              mn = signalNew;
              mnPos = x;
              lookForMax = false;
              return 0;
            }

            if (mxPos - maxtab.last[0] >= MAXTAB_MIN && mxPos - maxtab.last[0] < MAXTAB_MAX) {
              if (maxtab.last[0] > MAXTAB_CHECK_LEN) {
                slope1 = calculateSlope(buffer.sublist(mxPos - MAXTAB_BUFF, mxPos + 1));
                slope2 = calculateSlope(buffer.sublist(maxtab.last[0].toInt() - MAXTAB_BUFF, maxtab.last[0].toInt() + 1).toList());

                if (slope1.abs() < slope2.abs() * 0.5) {
                  mn = signalNew;
                  mnPos = x;
                  lookForMax = false;
                  return 0;
                }
              }
            }
          }

          maxtab.add([mxPos.toDouble(), mx]);
          mn = signalNew;
          mnPos = x;
          lookForMax = false;

          if (mx < 2.5 * delta) {
            delta = (0.85 * delta + 0.15 * mx) * 0.7;
            ignore = 0;
          } else {
            ignore++;
            if (ignore > 5) {
              delta = getMedian(buffer.sublist(x - MAX_BUFF_LEN, x + 1)) + meanAbsoluteDeviation(buffer.sublist(x - MAX_BUFF_LEN, x + 1));
              ignore = 0;
            }
          }
          cnt = 0;
          return mxPos;
        }
      } else {
        if (signalNew > mn + delta) {
          mx = signalNew;
          mxPos = x;
          lookForMax = true;
        }
        return 0;
      }
    }

    return 0;
  }

  int adjustPeaks(List<double> data, int oldInd) {
    maxVal = data.reduce(max);
    ind = data.indexOf(maxVal);
    return oldInd + ind;
  }

  List<int> getWidth(List<double> diffSeg, int startInd) {
    midVal = diffSeg.length / 2;
    final intDiff = <double>[];
    var y = <double>[];
    for (final sample in diffSeg) {
      if (sample > 0) {
        intDiff.add(1);
      } else {
        intDiff.add(0);
      }
    }

    for (var i = 0; i < intDiff.length - 1; i++) {
      y.add(intDiff[i + 1] - intDiff[i]);
    }

    final yMin = <int>[];
    final int yLength = y.length;
    for (var i = 0; i < yLength; i++) {
      if (y[i] > 0) {
        yMin.add(i + 1);
      }
    }

    final qCands = <int>[];
    final int yMinLength = yMin.length;
    for (var i = 0; i < yMinLength; i++) {
      if (yMin[i] < midVal) {
        qCands.add(i);
      }
    }

    if (qCands.isEmpty) {
      qCand = (midVal - midVal / 2).toInt();
    } else {
      qCand = yMin[qCands.reduce(max)];
    }

    var sCands = <int>[];
    for (var i = 0; i < yMinLength; i++) {
      if (yMin[i] > midVal) {
        sCands.add(i);
      }
    }

    if (sCands.isEmpty) {
      sCand = (midVal + midVal / 2).toInt();
    } else {
      sCand = yMin[sCands.reduce(min)];
    }

    qCand = qCand + startInd;
    sCand = sCand + startInd;

    return [qCand, sCand];
  }
}

class DwtDecompose {
  final decomHighpass = [
    -0.0322231006040427,
    -0.0126039672620378,
    0.0992195435768472,
    0.297857795605277,
    -0.803738751805916,
    0.497618667632015,
    0.0296355276459985,
    -0.0757657147892733,
  ];
  final decomLowpass = [
    -0.0757657147892733,
    -0.0296355276459985,
    0.497618667632015,
    0.803738751805916,
    0.297857795605277,
    -0.0992195435768472,
    -0.0126039672620378,
    0.0322231006040427,
  ];

  int ind = 0;
  List<double> bufferD = List<double>.filled(14, 0, growable: true);
  int filterLen = 8;
  double thresh = 0;

  //DwtDecompose();

  double dwtSample(List<double> ecgList, List<double> filter) {
    double val = 0;
    for (var i = 0; i < filterLen; i++) {
      val += ecgList[i] * filter[filterLen - i - 1];
    }
    return val;
  }

  List<double> decompose(double ecg) {
    double valApp = 0;
    double valDet = 0;
    var val = <double>[0, 0];
    if (ind < filterLen - 1) {
      bufferD[filterLen - 2 - ind] = ecg;
      bufferD[ind + filterLen - 1] = ecg;
      ind += 1;
    } else {
      if (ind.isEven) {
        valApp = dwtSample(bufferD.sublist(0, filterLen), decomLowpass);
        valDet = dwtSample(bufferD.sublist(0, filterLen), decomHighpass);
      }
      ind += 1;
      bufferD
        ..add(ecg)
        ..removeAt(0);
    }
    val[0] = valApp;
    val[1] = valDet;

    return [valApp, valDet];
  }

  double threshold(double x, double t) {
    double thresh = t.toDouble();

    double xNew = 0;
    double x1 = 0;
    if (x.abs() < thresh.abs()) {
      xNew = 0;
    } else {
      x1 = x.abs() - thresh;
      x1 = (x1 + x1.abs()) / 2;
      if (x < 0) {
        xNew = -1 * x1;
      } else {
        xNew = x1;
      }
    }
    return xNew;
  }
}

class DwtReconstruct {
  int filterLen = 8;
  double valRet_0 = 0;
  double valRet_1 = 0;
  List<double> filter = List.filled(8, 0);

  List<double> bufferR = List<double>.filled(8, 0, growable: true);
  List<double> bufferA = List<double>.filled(8, 0, growable: true);
  List<double> bufferD = List<double>.filled(8, 0, growable: true);
  List<double> recHighpass = [
    -0.0757657147892733,
    0.0296355276459985,
    0.497618667632015,
    -0.803738751805916,
    0.297857795605277,
    0.0992195435768472,
    -0.0126039672620378,
    -0.0322231006040427,
  ];
  List<double> recLowpass = [
    0.0322231006040427,
    -0.0126039672620378,
    -0.0992195435768472,
    0.297857795605277,
    0.803738751805916,
    0.497618667632015,
    -0.0296355276459985,
    -0.0757657147892733,
  ];

  //DwtReconstruct();

  double dwtSample(List<double> ecgList, List<double> filter) {
    double val = 0;
    for (var i = 0; i < filterLen; i++) {
      val += ecgList[i] * filter[filterLen - i - 1];
    }
    return val;
  }

  List<double> reconstruct(double coe, bool highpass) {
    if (highpass == true) {
      filter = recLowpass;
      bufferR = [...bufferA];
    } else {
      filter = recHighpass;
      bufferR = [...bufferD];
    }

    bufferR
      ..add(coe)
      ..removeAt(0);

    valRet_0 = dwtSample(bufferR.sublist(0, filterLen), filter);
    bufferR
      ..add(0)
      ..removeAt(0);
    valRet_1 = dwtSample(bufferR.sublist(0, filterLen), filter);

    if (highpass == true) {
      bufferA = [...bufferR];
    } else {
      bufferD = [...bufferR];
    }
    return [valRet_0, valRet_1];
  }
}

class DwtFilter {
  int level;
  double thres;

  int fs = AdaptiveHR.fs.toInt();
  int cnt = 0;
  int cutInt = 6;
  int cutInd = 0;
  int filterLen = 8;
  int waitInd = 0;
  List<double> newApp = [0, 0];
  List<double> reconstructedData = [0];
  List<int> minReq = [];
  List<double> results = [];
  List<double> app = [];
  List<double> det = [];
  List<DwtDecompose> wDec = [];
  List<DwtReconstruct> wRec = [];
  List<List<double>> newDet = [];
  List<List<double>> detWindow = [];
  List<double> reconApp = [];
  List<double> reconDet = [];
  List<int> window = [];
  List<double> cntApp = [];
  List<double> numApp = [];
  List<double> threshLevel = [];

  List<double> newDetL = [];
  List<double> absDetWindow = [];

  DwtFilter(this.thres, this.level) {
    wDec = List.generate(level, (_) => DwtDecompose());
    wRec = List.generate(level, (_) => DwtReconstruct());

    app = List<double>.filled(level, 0, growable: true);
    det = List<double>.filled(level, 0, growable: true);
    reconApp = List<double>.filled(level, 0, growable: true);
    reconDet = List<double>.filled(level, 0, growable: true);
    window = List.generate(level, (i) => (fs / pow(2, i + 1) * 2.5).floor());
    cntApp = List<double>.filled(level, 0, growable: true);
    numApp = List<double>.filled(level, 0, growable: true);
    threshLevel = List<double>.filled(level, 0, growable: true);

    newDet = List.generate(level, (_) => []);
    detWindow = List.generate(level, (_) => []);

    newDet[level - 1].add(0);

    var levelList = <int>[];
    for (var i = level; i > 0; i--) {
      levelList.add(i);
    }

    for (var j = 0; j < level; j++) {
      var numReq = 0;
      for (var i = levelList[j]; i > 0; i--) {
        numReq += pow(2, i - 1).toInt();
      }
      minReq.add(filterLen * numReq - 1);
    }
    minReq = minReq.reversed.toList();
    minReq.insert(0, 0);
  }

  double applyFilter(double sample) {
    for (var i = 0; i < level; i++) {
      if (((cnt - minReq[i] - 1) % pow(2, i).toInt()) == 0 && cnt >= minReq[i]) {
        results = wDec[i].decompose(sample);
        app[i] = results[0];
        det[i] = results[1];
        sample = app[i];
        if (i > 0) {
          if (i == 1) {
            newDet[i - 1].add(0);
          } else {
            detWindow[i - 1].add(det[i - 1]);
            if (detWindow[i - 1].length == window[i - 1]) {
              absDetWindow = [];
              for (var s = 0; s < detWindow[i - 1].length; s++) {
                absDetWindow.add(detWindow[i - 1][s].abs());
              }
              threshLevel[i - 1] = thres * absDetWindow.reduce(max);
              detWindow[i - 1].removeAt(0);
            }
            newDet[i - 1].add(wDec[i - 1].threshold(det[i - 1], threshLevel[i - 1]));
          }
        }
      }
    }

    var reconsData = <double>[];

    if ((cnt - minReq.last - 1) % pow(2, level) == 0 && cnt > minReq[level]) {
      reconsData = [0];
      reconsData[0] = app[level - 1];
      for (var i = level; i > 0; i--) {
        if (i == level) {
          detWindow[i - 1].add(det[i - 1]);
          if (detWindow[i - 1].length == window[i - 1]) {
            absDetWindow = [];
            for (var s = 0; s < detWindow[i - 1].length; s++) {
              absDetWindow.add(detWindow[i - 1][s].abs());
            }
            threshLevel[i - 1] = thres * absDetWindow.reduce(max);
            detWindow[i - 1].removeAt(0);
          }
          newDet[i - 1][0] = wDec[i - 1].threshold(det[i - 1], threshLevel[i - 1]);
        }
        newApp = reconsData;
        if (i == level) {
          newDetL = [...newDet[i - 1]];
          reconApp = wRec[i - 1].reconstruct(newApp[0], true);
          reconDet = wRec[i - 1].reconstruct(newDetL[0], false);
          newApp[0] = reconApp[0] + reconDet[0];
          newApp.add(reconApp[1] + reconDet[1]);
          reconsData = newApp;
        } else {
          reconsData = [];
          for (var k = 0; k < newApp.length; k++) {
            if (cntApp[i] >= cutInt) {
              cntApp[i] += 1;
              newDetL = [...newDet[i - 1]];
              newDet[i - 1].removeAt(0);
              reconApp = wRec[i - 1].reconstruct(newApp[k], true);
              reconDet = wRec[i - 1].reconstruct(newDetL[0], false);
              reconsData
                ..add(reconApp[0] + reconDet[0])
                ..add(reconApp[1] + reconDet[1]);
            } else {
              cntApp[i] += 1;
            }
          }
        }
      }
    }
    cnt += 1;

    if (level == 1) {
      reconstructedData = reconstructedData + reconsData;
      if (reconstructedData.length < 2) {
        return double.nan;
      } else {
        reconstructedData.removeAt(0);
        if (cutInd < 6) {
          cutInd += 1;
          return double.nan;
        } else {
          return reconstructedData[0];
        }
      }
    } else {
      if (cntApp[1] > 6) {
        reconstructedData = reconstructedData + reconsData;
        if (reconsData.isEmpty || reconsData.length == pow(2, level)) {
          if (waitInd > 0) {
            waitInd -= 1;
            return double.nan;
          } else {
            reconstructedData.removeAt(0);
            if (cutInd < 6) {
              cutInd += 1;
              return double.nan;
            } else {
              return reconstructedData[0];
            }
          }
        } else {
          waitInd = pow(2, level).toInt() - reconsData.length;
          waitInd -= 1;
          return double.nan;
        }
      } else {
        return double.nan;
      }
    }
  }
}

class SmoothingFilter {
  int ind = 0;
  List<double> buffer = List<double>.filled(17, 0, growable: true);
  final coe = [
    -0.0650154798761610,
    -0.0185758513931889,
    0.0216718266253870,
    0.0557275541795666,
    0.0835913312693498,
    0.105263157894737,
    0.120743034055728,
    0.130030959752322,
    0.133126934984520,
    0.130030959752322,
    0.120743034055728,
    0.105263157894737,
    0.0835913312693498,
    0.0557275541795666,
    0.0216718266253870,
    -0.0185758513931889,
    -0.0650154798761610
  ];

  //SmoothingFilter();

  double applySmoothing(double sample) {
    double result = 0;
    final int bufferLength = buffer.length;
    if (ind < bufferLength) {
      buffer[ind] = sample;
      if (ind == bufferLength - 1) {
        double sum = 0;

        for (var i = 0; i < ind + 1; i++) {
          sum += buffer[i] * coe[i];
        }
        result = sum;
      } else {
        result = sample;
      }
      ind += 1;
    } else {
      buffer
        ..removeAt(0)
        ..add(sample);
      double sum = 0;
      for (var i = 0; i < bufferLength; i++) {
        sum += buffer[i] * coe[i];
      }
      result = sum;
    }

    return result;
  }
}

class AdaptiveFilterRT {
  //double input = 0;
  double output = 0;
  double output2 = 0;
  double output3 = 0;
  int sample = -1;

  double output4 = 0;
  int outSample = -9;

  int rInd = 0;
  int rInd2 = 0;
  int rCand = 0;
  int eraseCnt = 0;

  double delta = 0;
  double newSig = 0;

  int nanCnt = 0;
  SmoothingFilter svFilter = SmoothingFilter();
  DwtFilter waveletFilter = DwtFilter(0.1, 3);
  ExcludePeaks removeR = ExcludePeaks();
  DetectPeaks detectR = DetectPeaks(0);

  List<double> dataList = [];
  List<double> newData = [];
  List<double> movingFiltered = [];
  List<double> lpFiltered = [];
  List<int> qsCand = [];
  List<double> thresholds = [];
  List<double> diffSeg = [];
  List<double> dataOrg = [];
  List<double> smoothedData = [];
  List<double> lpFiltered2 = [];
  List<double> diffData = [];
  List<double> diffData2 = [];
  List<double> seg = [];

  bool searchBool = false;
  bool removeBool = false;
  bool filterBool = false;

  var maxDuration = 0;

  double hr = 0;
  int rPrev = 0;
  int rDetected = 0;
  int fs = AdaptiveHR.fs.toInt();

  List<double> applyFilter(double input) {
    dataList.add(input);
    output = waveletFilter.applyFilter(input);
    if (output.isNaN) {
      nanCnt += 1;
      return [0, 0];
    } else {
      sample++;
      outSample++;
      dataOrg.add(dataList[sample]);
      newData.add(output);
      output2 = svFilter.applySmoothing(output);
      smoothedData.add(output2);
      output3 = removeR.movingAvFilt(output2);
      movingFiltered.add(output3);
      output4 = removeR.lowpassFilt(output3);
      lpFiltered.add(output4);
      if (lpFiltered.length > 8) {
        lpFiltered2.add(output4);
        if (lpFiltered2.length == 500) {
          delta = detectR.getMedian(lpFiltered2.sublist(0, 500)) + detectR.meanAbsoluteDeviation(lpFiltered2.sublist(0, 500));
          detectR.inputDelta(delta);
        }

        if (lpFiltered2.length > 500) {
          // 잠재적 리스크: outSample의 초기 값이 -9이므로 range error를 만들 수 있음
          rInd = detectR.findPeaks(lpFiltered2[outSample]);
        }

        if (rInd > 0) {
          searchBool = true;
          rCand = rInd + 500;
        } else {
          hr = 0;
        }
        if (searchBool) {
          if (outSample - rCand > 41) {
            searchBool = false;
            rInd2 = detectR.adjustPeaks(dataOrg.sublist(rCand - 37, rCand), rCand - 37);

            rDetected = rInd2;
            if (rPrev > 0) {
              hr = 60 / ((rDetected - rPrev) / fs);
            }
            rPrev = rDetected;

            diffSeg = [];
            for (var i = rInd2 + 8 - 30; i < rInd2 + 8 + 30; i++) {
              diffSeg.add(smoothedData[i + 1] - smoothedData[i]);
            }
            qsCand = detectR.getWidth(diffSeg, rInd2 + 8 - 30);

            for (var i = qsCand[0]; i < qsCand[1] + 1; i++) {
              smoothedData[i] = dataOrg[i - 8];
              diffData[i - 8] = 0;
            }

            if (qsCand[0] - 8 >= 500) {
              thresholds = removeR.calculateThreshold(diffData, qsCand[0] - 8 - 500, qsCand[0] - 8);
              removeBool = removeR.removeResidualQrs(diffData[qsCand[0] - 8 - 1], thresholds[0], thresholds[1]);

              if (removeBool) {
                for (var i = qsCand[0] - 4; i < qsCand[0]; i++) {
                  smoothedData[i] = dataOrg[i - 8];
                }
                removeBool = false;
                qsCand[0] = qsCand[0] - 4;
              }

              thresholds = removeR.calculateThreshold(diffData, qsCand[1] - 8 - 500, qsCand[1] - 8);
              removeBool = removeR.removeResidualQrs(diffData[qsCand[1] - 8 - 1], thresholds[0], thresholds[1]);

              if (removeBool) {
                for (var i = qsCand[1] + 1; i < qsCand[1] + 5; i++) {
                  smoothedData[i] = dataOrg[i - 8];
                }
                removeBool = false;
                qsCand[1] = qsCand[1] + 4;
              }

              var sm = 0;
              for (var i = qsCand[0] - 1; i < qsCand[0] + 2; i++) {
                seg = smoothedData.sublist(qsCand[0] - 2 + sm, qsCand[0] + sm + 1);
                newSig = 1 / 3 * seg.reduce((a, b) => a + b);
                smoothedData[i] = newSig;
                sm++;
              }

              sm = 0;
              for (var i = qsCand[1] - 1; i < qsCand[1] + 2; i++) {
                seg = smoothedData.sublist(qsCand[1] - 2 + sm, qsCand[1] + sm + 1);
                newSig = 1 / 3 * seg.reduce((a, b) => a + b);
                smoothedData[i] = newSig;
                sm++;
              }
              filterBool = true;
            }
          }
        }
        diffData.add(dataOrg[outSample] - smoothedData[sample]);
      }

      if (sample > 150) {
        if (filterBool) {
          return [smoothedData[sample + 8 - 143], hr];
        } else {
          return [dataOrg[sample - 143], 0];
        }
      } else {
        return [0, 0];
      }
    }
  }
}
